"""
Tests for the `edx_name_affirmation` Python API.
"""

import ddt

from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.test import TestCase

from edx_name_affirmation.api import (
    create_verified_name,
    create_verified_name_config,
    get_verified_name,
    get_verified_name_history,
    should_use_verified_name_for_certs,
    update_verification_attempt_id,
    update_verified_name_status
)
from edx_name_affirmation.exceptions import (
    VerifiedNameAttemptIdNotGiven,
    VerifiedNameDoesNotExist,
    VerifiedNameEmptyString,
    VerifiedNameMultipleAttemptIds
)
from edx_name_affirmation.models import VerifiedName, VerifiedNameConfig
from edx_name_affirmation.statuses import VerifiedNameStatus

User = get_user_model()


@ddt.ddt
class TestVerifiedNameAPI(TestCase):
    """
    Tests for the VerifiedName API.
    """
    VERIFIED_NAME = 'Jonathan Doe'
    PROFILE_NAME = 'Jon Doe'
    VERIFICATION_ATTEMPT_ID = 123
    PROCTORED_EXAM_ATTEMPT_ID = 456

    def setUp(self):
        super().setUp()
        self.user = User(username='jondoe', email='jondoe@test.com')
        self.user.save()
        # Create a fresh config with default values
        VerifiedNameConfig.objects.create(user=self.user)

    def tearDown(self):
        super().tearDown()
        cache.clear()

    def test_create_verified_name_defaults(self):
        """
        Test to create a verified name with default values.
        """
        verified_name_obj = self._create_verified_name()

        self.assertEqual(verified_name_obj.user, self.user)
        self.assertIsNone(verified_name_obj.verification_attempt_id)
        self.assertIsNone(verified_name_obj.proctored_exam_attempt_id)
        self.assertEqual(verified_name_obj.status, VerifiedNameStatus.PENDING.value)

    @ddt.data(
        (123, None, VerifiedNameStatus.APPROVED),
        (None, 456, VerifiedNameStatus.SUBMITTED),
    )
    @ddt.unpack
    def test_create_verified_name_with_optional_arguments(
        self, verification_attempt_id, proctored_exam_attempt_id, status,
    ):
        """
        Test to create a verified name with optional arguments supplied.
        """
        verified_name_obj = self._create_verified_name(
            verification_attempt_id, proctored_exam_attempt_id, status,
        )

        self.assertEqual(verified_name_obj.verification_attempt_id, verification_attempt_id)
        self.assertEqual(verified_name_obj.proctored_exam_attempt_id, proctored_exam_attempt_id)
        self.assertEqual(verified_name_obj.status, status.value)

    def test_create_verified_name_two_ids(self):
        """
        Test that a verified name cannot be created with both a verification_attempt_id
        and a proctored_exam_attempt_id.
        """
        with self.assertRaises(VerifiedNameMultipleAttemptIds):
            create_verified_name(
                self.user,
                self.VERIFIED_NAME,
                self.PROFILE_NAME,
                self.VERIFICATION_ATTEMPT_ID,
                self.PROCTORED_EXAM_ATTEMPT_ID,
            )

    @ddt.data(
        ('', PROFILE_NAME),
        (VERIFIED_NAME, ''),
    )
    @ddt.unpack
    def test_create_verified_name_empty_string(self, verified_name, profile_name):
        """
        Test that an empty verified_name or profile_name will raise an exception.
        """
        field = ''
        if verified_name == '':
            field = 'verified_name'
        elif profile_name == '':
            field = 'profile_name'

        with self.assertRaises(VerifiedNameEmptyString) as context:
            create_verified_name(self.user, verified_name, profile_name)

        self.assertEqual(
            str(context.exception),
            'Attempted to create VerifiedName for user_id={user_id}, but {field} was '
            'empty.'.format(field=field, user_id=self.user.id),
        )

    def test_get_verified_name_most_recent(self):
        """
        Test to get the most recent verified name.
        """
        create_verified_name(self.user, 'old verified name', 'old profile name')
        self._create_verified_name()

        verified_name_obj = get_verified_name(self.user)

        self.assertEqual(verified_name_obj.verified_name, self.VERIFIED_NAME)
        self.assertEqual(verified_name_obj.profile_name, self.PROFILE_NAME)

    def test_get_verified_name_only_verified(self):
        """
        Test that VerifiedName entries with status != approved are ignored if is_verified
        argument is set to True.
        """
        self._create_verified_name(status=VerifiedNameStatus.APPROVED)
        create_verified_name(self.user, 'unverified name', 'unverified profile name')

        verified_name_obj = get_verified_name(self.user, True)

        self.assertEqual(verified_name_obj.verified_name, self.VERIFIED_NAME)
        self.assertEqual(verified_name_obj.profile_name, self.PROFILE_NAME)

    @ddt.data(False, True)
    def test_get_verified_name_none_exist(self, check_is_verified):
        """
        Test that None returns if there are no VerifiedName entries. If the `is_verified`
        flag is set to True, and there are only non-verified entries, we should get the
        same result.
        """
        if check_is_verified:
            self._create_verified_name()
            verified_name_obj = get_verified_name(self.user, True)
        else:
            verified_name_obj = get_verified_name(self.user)

        self.assertIsNone(verified_name_obj)

    @ddt.data(
        ([VerifiedNameStatus.PENDING], VerifiedNameStatus.DENIED),
        ([VerifiedNameStatus.DENIED, VerifiedNameStatus.PENDING], VerifiedNameStatus.APPROVED),
        (['invalid status'], VerifiedNameStatus.PENDING)
    )
    @ddt.unpack
    def test_get_verified_name_exclude(self, statuses_to_exclude, expected_status):
        """
        Exclude verified names by status when getting the most recent verified name.
        """
        self._create_verified_name(status=VerifiedNameStatus.APPROVED)
        self._create_verified_name(status=VerifiedNameStatus.DENIED)
        self._create_verified_name(status=VerifiedNameStatus.PENDING)

        verified_name_obj = get_verified_name(self.user, False, statuses_to_exclude)

        self.assertEqual(verified_name_obj.status, expected_status)

    @ddt.data(False, True)
    def test_get_verified_name_ignore_exlude_parameter(self, is_verified):
        """
        If `is_verified` is True, ignore the `statuses_to_exclude` parameter.
        """
        self._create_verified_name(status=VerifiedNameStatus.APPROVED)

        verified_name_obj = get_verified_name(self.user, is_verified, [VerifiedNameStatus.APPROVED])

        if is_verified:
            self.assertEqual(verified_name_obj.verified_name, self.VERIFIED_NAME)
        else:
            self.assertIsNone(verified_name_obj)

    def test_get_verified_name_history(self):
        """
        Test that get_verified_name_history returns all of the user's VerifiedNames
        ordered by most recently created.
        """
        verified_name_first = self._create_verified_name()
        verified_name_second = self._create_verified_name()

        verified_name_qs = get_verified_name_history(self.user)
        self.assertEqual(verified_name_qs[1].id, verified_name_first.id)
        self.assertEqual(verified_name_qs[0].id, verified_name_second.id)

    def test_update_verification_attempt_id(self):
        """
        Test that the most recent VerifiedName is updated with a verification_attempt_id if
        it does not already have one.
        """
        first_verified_name_id = self._create_verified_name().id
        second_verified_name_id = self._create_verified_name().id

        update_verification_attempt_id(self.user, self.VERIFICATION_ATTEMPT_ID)

        first_verified_name_obj = VerifiedName.objects.get(id=first_verified_name_id)
        second_verified_name_obj = VerifiedName.objects.get(id=second_verified_name_id)

        self.assertIsNone(first_verified_name_obj.verification_attempt_id)
        self.assertEqual(second_verified_name_obj.verification_attempt_id, self.VERIFICATION_ATTEMPT_ID)

    @ddt.data(
        (VERIFICATION_ATTEMPT_ID, None),
        (None, PROCTORED_EXAM_ATTEMPT_ID),
    )
    @ddt.unpack
    def test_update_verification_attempt_id_already_exists(
        self, verification_attempt_id, proctored_exam_attempt_id,
    ):
        """
        Test that if the most recent VerifiedName already has a linked verification or
        proctored exam attempt, a new VerifiedName will be created when updating the
        `verification_attempt_id`.
        """
        self._create_verified_name(
            verification_attempt_id=verification_attempt_id,
            proctored_exam_attempt_id=proctored_exam_attempt_id,
        )
        update_verification_attempt_id(self.user, 789)
        verified_name_qs = VerifiedName.objects.all()
        self.assertEqual(len(verified_name_qs), 2)

    def test_update_verification_attempt_id_none_exist(self):
        """
        Test that if the user does not have an existing VerifiedName,
        `update_verification_attempt_id` will raise an exception.
        """
        with self.assertRaises(VerifiedNameDoesNotExist):
            update_verification_attempt_id(self.user, self.VERIFICATION_ATTEMPT_ID)

    @ddt.data(
        (VERIFICATION_ATTEMPT_ID, None),
        (None, PROCTORED_EXAM_ATTEMPT_ID)
    )
    @ddt.unpack
    def test_update_is_verified_status(
        self, verification_attempt_id, proctored_exam_attempt_id,
    ):
        """
        Test that VerifiedName status can be updated with a given attempt ID.
        """
        self._create_verified_name(verification_attempt_id, proctored_exam_attempt_id)
        update_verified_name_status(
            self.user, VerifiedNameStatus.DENIED, verification_attempt_id, proctored_exam_attempt_id,
        )
        verified_name_obj = get_verified_name(self.user)
        self.assertEqual(VerifiedNameStatus.DENIED.value, verified_name_obj.status)

    def test_update_is_verified_no_attempt_id(self):
        """
        Test that `update_is_verified_by_attempt_id` will raise an exception with no attempt
        ID given.
        """
        with self.assertRaises(VerifiedNameAttemptIdNotGiven):
            update_verified_name_status(self.user, True)

    def test_update_is_verified_multiple_attempt_ids(self):
        """
        Test that `update_is_verified_by_attempt_id` will raise an exception with multiple attempt
        IDs given.
        """
        with self.assertRaises(VerifiedNameMultipleAttemptIds):
            update_verified_name_status(
                self.user, True, self.VERIFICATION_ATTEMPT_ID, self.PROCTORED_EXAM_ATTEMPT_ID,
            )

    def test_update_is_verified_does_not_exist(self):
        """
        Test that `update_is_verified_by_attempt_id` will raise an exception if a VerifiedName does
        not exist for the attempt ID given.
        """
        with self.assertRaises(VerifiedNameDoesNotExist):
            update_verified_name_status(self.user, True, self.VERIFICATION_ATTEMPT_ID)

    def _create_verified_name(
        self, verification_attempt_id=None, proctored_exam_attempt_id=None, status=VerifiedNameStatus.PENDING,
    ):
        """
        Util to create and return a VerifiedName with default names.
        """
        create_verified_name(
            self.user, self.VERIFIED_NAME, self.PROFILE_NAME, verification_attempt_id,
            proctored_exam_attempt_id, status
        )
        return get_verified_name(self.user)

    @ddt.data(
        (True, True),
        (True, False),
        (False, False),
    )
    @ddt.unpack
    def test_should_use_verified_name_for_certs(self, create_config, expected_value):
        """
        Test that the correct config value is returned from `should_use_verified_name_for_certs`
        """
        if create_config:
            VerifiedNameConfig.objects.create(user=self.user, use_verified_name_for_certs=expected_value)

        # create config for other user, make sure it is the opposite of the expected value.
        # we want to add an additional config to make sure that caching by user works properly
        other_user = User(username='bobsmith', email='bobsmith@test.com')
        other_user.save()
        VerifiedNameConfig.objects.create(user=other_user, use_verified_name_for_certs=not expected_value)

        should_use_for_certs = should_use_verified_name_for_certs(self.user)
        self.assertEqual(should_use_for_certs, expected_value)

    def test_create_verified_name_config(self):
        """
        Test that verified name config is created and updated successfully
        """
        create_verified_name_config(self.user, use_verified_name_for_certs=True)

        # check that new record was created
        config_obj = VerifiedNameConfig.current(self.user)
        self.assertTrue(config_obj.use_verified_name_for_certs)
        self.assertEqual(config_obj.user, self.user)

    def test_create_verified_name_config_no_overwrite(self):
        """
        Test that if a field is set to True, it will not be overridden by False
        if not specified when the config is updated
        """
        create_verified_name_config(self.user, use_verified_name_for_certs=True)
        create_verified_name_config(self.user)
        self.assertTrue(should_use_verified_name_for_certs(self.user))



# pylint: disable=logging-format-interpolation, unused-argument
"""
Name affirmation celery tasks
"""

import logging

from celery import shared_task
from edx_django_utils.monitoring import set_code_owner_attribute

from django.contrib.auth import get_user_model
from django.db.models import Q

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.statuses import VerifiedNameStatus

User = get_user_model()

log = logging.getLogger(__name__)

DEFAULT_RETRY_SECONDS = 30
MAX_RETRIES = 3


@shared_task(
    bind=True, autoretry_for=(Exception,), default_retry_delay=DEFAULT_RETRY_SECONDS, max_retries=MAX_RETRIES,
)
@set_code_owner_attribute
def idv_update_verified_name_task(self, attempt_id, user_id, name_affirmation_status, photo_id_name, full_name):
    """
    Celery task for updating a verified name based on an IDV attempt
    """
    log.info('VerifiedName: idv_update_verified_name triggering Celery task started for user %(user_id)s '
             'with attempt_id %(attempt_id)s and status %(status)s',
             {
                'user_id': user_id,
                'attempt_id': attempt_id,
                'status': name_affirmation_status
             }
             )
    # get all verified names that are either not associated with an IDV attempt, or
    # only associated with the IDV attempt for which we received an update. We do not
    # want to grab all verified names for the same user and name combination, because
    # some of those records may already be associated with a different IDV attempt.
    verified_names = VerifiedName.objects.filter(
        (Q(verification_attempt_id=attempt_id) | Q(verification_attempt_id__isnull=True))
        & Q(user__id=user_id)
        & Q(verified_name=photo_id_name)
    ).order_by('-created')
    if verified_names:
        # if there are VerifiedName objects, we want to update existing entries
        # for each attempt with no attempt id (either proctoring or idv), update attempt id
        updated_for_attempt_id = verified_names.filter(
            proctored_exam_attempt_id=None,
            verification_attempt_id=None
        ).update(verification_attempt_id=attempt_id)

        if updated_for_attempt_id:
            log.info(
                'Updated VerifiedNames for user={user_id} to verification_attempt_id={attempt_id}'.format(
                    user_id=user_id,
                    attempt_id=attempt_id,
                )
            )

        # then for all matching attempt ids, update the status
        verified_name_qs = verified_names.filter(
            verification_attempt_id=attempt_id,
            proctored_exam_attempt_id=None
        )

        # Individually update to ensure that post_save signals send
        for verified_name_obj in verified_name_qs:
            verified_name_obj.status = name_affirmation_status
            verified_name_obj.save()

        log.info(
            'Updated VerifiedNames for user={user_id} with verification_attempt_id={attempt_id} to '
            'have status={status}'.format(
                user_id=user_id,
                attempt_id=attempt_id,
                status=name_affirmation_status
            )
        )
    else:
        # otherwise if there are no entries, we want to create one.
        user = User.objects.get(id=user_id)
        verified_name = VerifiedName.objects.create(
            user=user,
            verified_name=photo_id_name,
            profile_name=full_name,
            verification_attempt_id=attempt_id,
            status=name_affirmation_status,
        )
        log.error(
            'Created VerifiedName for user={user_id} to have status={status} '
            'and verification_attempt_id={attempt_id}, because no matching '
            'attempt_id or verified_name were found.'.format(
                user_id=user_id,
                attempt_id=attempt_id,
                status=verified_name.status
            )
        )


@shared_task(
    bind=True, autoretry_for=(Exception,), default_retry_delay=DEFAULT_RETRY_SECONDS, max_retries=MAX_RETRIES,
)
@set_code_owner_attribute
def proctoring_update_verified_name_task(
    self,
    attempt_id,
    user_id,
    name_affirmation_status,
    full_name,
    profile_name
):
    """
    Celery task for updating a verified name based on a proctoring attempt
    """

    approved_verified_name = VerifiedName.objects.filter(
        user__id=user_id,
        status=VerifiedNameStatus.APPROVED
    ).order_by('-created').first()

    verified_name_for_exam = VerifiedName.objects.filter(
        user__id=user_id,
        proctored_exam_attempt_id=attempt_id
    ).order_by('-created').first()

    # check if approved VerifiedName already exists for the user, and skip
    # update if no VerifiedName has already been created for this specific exam
    if approved_verified_name and not verified_name_for_exam:
        is_full_name_approved = approved_verified_name.verified_name == full_name
        if not is_full_name_approved:
            log.warning(
                'Full name for proctored_exam_attempt_id={attempt_id} is not equal '
                'to the most recent verified name verified_name_id={name_id}.'.format(
                    attempt_id=attempt_id,
                    name_id=approved_verified_name.id
                )
            )
        return

    if verified_name_for_exam:
        verified_name_for_exam.status = name_affirmation_status
        verified_name_for_exam.save()
        log.info(
            'Updated VerifiedName for user={user_id} with proctored_exam_attempt_id={attempt_id} '
            'to have status={status}'.format(
                user_id=user_id,
                attempt_id=attempt_id,
                status=name_affirmation_status
            )
        )
    else:
        if full_name and profile_name:
            # if they do not already have an approved VerifiedName, create one
            user = User.objects.get(id=user_id)
            VerifiedName.objects.create(
                user=user,
                verified_name=full_name,
                proctored_exam_attempt_id=attempt_id,
                status=name_affirmation_status,
                profile_name=profile_name
            )
            log.info(
                'Created VerifiedName for user={user_id} to have status={status} '
                'and proctored_exam_attempt_id={attempt_id}'.format(
                    user_id=user_id,
                    attempt_id=attempt_id,
                    status=name_affirmation_status
                )
            )
        else:
            log.error(
                'Cannot create VerifiedName for user={user_id} for proctored_exam_attempt_id={attempt_id} '
                'because neither profile name nor full name were provided'.format(
                    user_id=user_id,
                    attempt_id=attempt_id,
                )
            )


@shared_task(
    bind=True, autoretry_for=(Exception,), default_retry_delay=DEFAULT_RETRY_SECONDS, max_retries=MAX_RETRIES,
)
@set_code_owner_attribute
def delete_verified_name_task(self, idv_attempt_id, proctoring_attempt_id):
    """
    Celery task to delete a verified name based on an idv or proctoring attempt
    """
    # this case shouldn't happen, but should log as an error in case
    if (idv_attempt_id and proctoring_attempt_id) or (not idv_attempt_id and not proctoring_attempt_id):
        log.error(
            'A maximum of one attempt id should be provided for either a proctored exam attempt or IDV attempt.'
        )
        return

    log_message = {'field_name': '', 'attempt_id': ''}

    if idv_attempt_id:
        verified_names = VerifiedName.objects.filter(verification_attempt_id=idv_attempt_id)
        log_message['field_name'] = 'verification_attempt_id'
        log_message['attempt_id'] = idv_attempt_id
    else:
        verified_names = VerifiedName.objects.filter(proctored_exam_attempt_id=proctoring_attempt_id)
        log_message['field_name'] = 'proctored_exam_attempt_id'
        log_message['attempt_id'] = proctoring_attempt_id

    if verified_names:
        log.info(
            'Deleting {num_names} VerifiedName(s) associated with {field_name}='
            '{verification_attempt_id}'.format(
                num_names=len(verified_names),
                field_name=log_message['field_name'],
                verification_attempt_id=log_message['attempt_id'],
            )
        )
        verified_names.delete()

    log.info(
        'No VerifiedNames deleted because no VerifiedNames were associated with the provided attempt ID.'
    )



# Generated by Django 2.2.24 on 2021-08-13 12:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edx_name_affirmation', '0002_verifiednameconfig'),
    ]

    operations = [
        migrations.AddField(
            model_name='verifiedname',
            name='status',
            field=models.CharField(choices=[('pending', 'pending'), ('submitted', 'submitted'), ('approved', 'approved'), ('denied', 'denied')], default='pending', max_length=32),
        ),
    ]



# Generated by Django 2.2.24 on 2021-08-30 20:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edx_name_affirmation', '0005_remove_verifiedname_is_verified'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verifiedname',
            name='proctored_exam_attempt_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='verifiedname',
            name='verification_attempt_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



"""
Name Change Validator
"""

import re
from difflib import SequenceMatcher


class NameChangeValidator:
    """
    Class used to validate name changes
    """

    # maximum number of name changes allowed before verification should be triggered
    MAX_NUM_NAME_CHANGES = 2

    def __init__(self, old_names_list, num_certs, old_name, new_name):
        """
        Class initializer
        """
        self.old_names_list = old_names_list
        self.num_certs = num_certs
        self.old_name = old_name
        self.new_name = new_name

    def _validate_spaces(self):
        """
        Validate spaces in a new name

        Returns a boolean representing if spaces within string follow current rules
        """
        contains_multiple_spaces = bool(re.search(r' {2,}', self.new_name))
        if contains_multiple_spaces:
            return False
        return True

    def _validate_string_changes(self):
        """
        Validate any changes made from the old name to the new name

        Returns a boolean representing if the changes follow current rules
        Edits are considered invalid if:
            * Two or more spaces occur in a row
            * More than one non-space character is added/removed/replaced (exception for if a space
              is added on either side of the non-space character)
        """
        modifications = 0
        # get differences between old name and new name
        sequence = SequenceMatcher(lambda x: x == " ", self.old_name, self.new_name)
        for tag, i1, i2, j1, j2 in sequence.get_opcodes():
            # if there is more than one sequence in the string that has been modified, edits are invalid
            if modifications > 1:
                return False

            # if tag is anything other than equal, increase modifications
            if tag != 'equal':
                modifications += 1

                # determine which piece has been modified
                old_name_substring = self.old_name[i1:i2]
                new_name_substring = self.new_name[j1:j2]
                modified_substring = (
                    old_name_substring
                    if len(old_name_substring) > len(new_name_substring)
                    else new_name_substring
                )
                is_valid = bool(re.search(r'^\s?(\S\s?)?$', modified_substring))
                if not is_valid:
                    return False

        return True

    def _validate_old_name_changes(self):
        """
        Validate that a user has not changed their name more than the maximum number allowed
        """
        return len(self.old_names_list) < self.MAX_NUM_NAME_CHANGES

    def _validate_num_certs(self):
        """
        Validate that the user does not have any certificates
        """
        return self.num_certs == 0

    def validate(self):
        """
        Return a boolean representing if the edits to a name are valid and follow the current rules for validation
        """
        return (
            self._validate_num_certs()  # if a user has no certs, changes will always be considered valid
            or (
                self._validate_spaces()
                and self._validate_string_changes()
                and self._validate_old_name_changes()
            )
        )



"""
Database models for edx_name_affirmation.
"""

from config_models.models import ConfigurationModel
from model_utils.models import TimeStampedModel
from simple_history.models import HistoricalRecords

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from edx_name_affirmation.statuses import VerifiedNameStatus

try:
    from lms.djangoapps.verify_student.models import SoftwareSecurePhotoVerification
except ImportError:
    SoftwareSecurePhotoVerification = None

User = get_user_model()


class VerifiedName(TimeStampedModel):
    """
    This model represents a verified name for a user, with a link to the source
    through `verification_attempt_id` or `proctored_exam_attempt_id` if applicable.

    .. pii: Contains name fields.
    .. pii_types: name
    .. pii_retirement: local_api
    """
    user = models.ForeignKey(User, db_index=True, on_delete=models.CASCADE)
    verified_name = models.CharField(max_length=255, db_index=True)

    # Snapshot of the user's UserProfile `name` upon creation
    profile_name = models.CharField(max_length=255, null=True)

    # Reference to an external ID verification or proctored exam attempt
    verification_attempt_id = models.PositiveIntegerField(null=True, blank=True)
    proctored_exam_attempt_id = models.PositiveIntegerField(null=True, blank=True)

    status = models.CharField(
        max_length=32,
        choices=[(st.value, st.value) for st in VerifiedNameStatus],
        default=VerifiedNameStatus.PENDING.value,
    )
    history = HistoricalRecords()

    class Meta:
        """ Meta class for this Django model """
        db_table = 'nameaffirmation_verifiedname'
        verbose_name = 'verified name'

    @property
    def verification_attempt_status(self):
        "Returns the status associated with its SoftwareSecurePhotoVerification with verification_attempt_id if any."

        if not self.verification_attempt_id or not SoftwareSecurePhotoVerification:
            return None

        try:
            verification = SoftwareSecurePhotoVerification.objects.get(id=self.verification_attempt_id)
            return verification.status

        except ObjectDoesNotExist:
            return None


class VerifiedNameConfig(ConfigurationModel):
    """
    This model provides various configuration fields for users regarding their
    verified name.
    .. no_pii: This model has no PII.
    """
    KEY_FIELDS = ('user',)

    user = models.ForeignKey(User, db_index=True, on_delete=models.CASCADE, related_name='verified_name_config')
    use_verified_name_for_certs = models.BooleanField(default=False)

    class Meta:
        """ Meta class for this Django model """
        db_table = 'nameaffirmation_verifiednameconfig'
        verbose_name = 'verified name config'



# Generated by Django 2.2.24 on 2021-08-16 09:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edx_name_affirmation', '0003_verifiedname_status'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verifiedname',
            name='is_verified',
            field=models.BooleanField(default=False, null=True),
        ),
    ]



"""Defines serializers used by the Name Affirmation API"""

import re

from rest_framework import serializers

from django.contrib.auth import get_user_model

from edx_name_affirmation.models import VerifiedName, VerifiedNameConfig

User = get_user_model()


class VerifiedNameSerializer(serializers.ModelSerializer):
    """
    Serializer for the VerifiedName Model.
    """
    username = serializers.CharField(source="user.username")
    verified_name = serializers.CharField(required=True)
    profile_name = serializers.CharField(required=True)
    verification_attempt_id = serializers.IntegerField(required=False, allow_null=True)
    verification_attempt_status = serializers.CharField(required=False, allow_null=True)
    proctored_exam_attempt_id = serializers.IntegerField(required=False, allow_null=True)
    status = serializers.CharField(required=False, allow_null=True)

    class Meta:
        """
        Meta Class
        """
        model = VerifiedName

        fields = (
            "id", "created", "username", "verified_name", "profile_name", "verification_attempt_id",
            "verification_attempt_status", "proctored_exam_attempt_id", "status"
        )

    def validate_verified_name(self, verified_name):
        if self._contains_html(verified_name):
            raise serializers.ValidationError('Name cannot contain the following characters: < >')
        if self._contains_url(verified_name):
            raise serializers.ValidationError('Name cannot contain a URL')

    def _contains_html(self, string):
        """
        Validator method to check whether a string contains HTML tags
        """
        regex = re.compile('(<|>)', re.UNICODE)
        return bool(regex.search(string))

    def _contains_url(self, string):
        """
        Validator method to check whether a string contains a url
        """
        regex = re.findall(r'https|http?://(?:[-\w.]|(?:%[\da-fA-F]{2}))+', string)
        return bool(regex)


class UpdateVerifiedNameSerializer(VerifiedNameSerializer):
    """
    Serializer for updates to the VerifiedName Model.
    """
    username = serializers.CharField(source='user.username', required=True)
    verified_name = serializers.CharField(required=False)
    profile_name = serializers.CharField(required=False)
    status = serializers.CharField(required=True)


class VerifiedNameConfigSerializer(serializers.ModelSerializer):
    """
    Serializer for the VerifiedNameConfig Model.
    """
    username = serializers.CharField(source="user.username")
    use_verified_name_for_certs = serializers.BooleanField(required=False, allow_null=True)

    class Meta:
        """
        Meta Class
        """
        model = VerifiedNameConfig

        fields = ("change_date", "username", "use_verified_name_for_certs")



"""
Tests for Name Affirmation models
"""
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.statuses import VerifiedNameStatus

User = get_user_model()


class VerifiedNameModelTests(TestCase):
    """
    Test suite for the VerifiedName models
    """
    def setUp(self):
        self.idv_attempt_id = 34455
        self.idv_attempt_status = 'submitted'
        self.idv_attempt_id_notfound = 404
        self.verified_name = 'Test Tester'
        self.user = User.objects.create(username='modelTester', email='model@tester.com')
        self.verified_name = VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            status=VerifiedNameStatus.SUBMITTED,
        )
        return super().setUp()

    def test_histories(self):
        """
        Test the model history is recording records as expected
        """

        verified_name_history = self.verified_name.history.all().order_by('history_date')
        assert len(verified_name_history) == 1
        self.verified_name.status = VerifiedNameStatus.APPROVED
        self.verified_name.verification_attempt_id = self.idv_attempt_id
        self.verified_name.save()
        verified_name_history = self.verified_name.history.all().order_by('history_date')
        assert len(verified_name_history) == 2

        first_history_record = verified_name_history[0]
        assert first_history_record.status == VerifiedNameStatus.SUBMITTED
        assert first_history_record.verification_attempt_id is None

        second_history_record = verified_name_history[1]
        assert second_history_record.status == VerifiedNameStatus.APPROVED
        assert second_history_record.verification_attempt_id == self.idv_attempt_id

    @patch('edx_name_affirmation.models.SoftwareSecurePhotoVerification')
    def test_verification_status(self, sspv_mock):
        """
        Test the model history is recording records as expected
        """

        idv_attempt_id_notfound_status = None

        sspv_mock.objects.get = self._mocked_model_get

        self.verified_name.verification_attempt_id = self.idv_attempt_id_notfound
        assert self.verified_name.verification_attempt_status is idv_attempt_id_notfound_status

        self.verified_name.verification_attempt_id = self.idv_attempt_id
        assert self.verified_name.verification_attempt_status is self.idv_attempt_status

    # Helper methods

    def _obj(self, dictionary):
        "Helper method to turn a dict into an object. Used to mock below."

        return type('obj', (object,), dictionary)

    def _mocked_model_get(self, id):  # pylint: disable=redefined-builtin
        "Helper method to mock the behavior of SoftwareSecurePhotoVerification model. Used to mock below."
        if id == self.idv_attempt_id_notfound:
            raise ObjectDoesNotExist

        if id == self.idv_attempt_id:
            return self._obj({'status': self.idv_attempt_status})

        return self._obj({'status': None})



"""
Name Affirmation signal handlers
"""

import logging

from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch.dispatcher import receiver

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.signals import VERIFIED_NAME_APPROVED
from edx_name_affirmation.statuses import VerifiedNameStatus
from edx_name_affirmation.tasks import (
    delete_verified_name_task,
    idv_update_verified_name_task,
    proctoring_update_verified_name_task
)

User = get_user_model()

log = logging.getLogger(__name__)


@receiver(post_save, sender=VerifiedName)
def verified_name_approved(sender, instance, **kwargs):  # pylint: disable=unused-argument
    """
    Emit a signal when a verified name's status is updated to "approved".
    """
    if instance.status == VerifiedNameStatus.APPROVED:
        VERIFIED_NAME_APPROVED.send(
          sender='name_affirmation',
          user_id=instance.user.id,
          profile_name=instance.profile_name
        )


def idv_attempt_handler(attempt_id, user_id, status, photo_id_name, full_name, **kwargs):
    """
    Receiver for IDV attempt updates

    Args:
        attempt_id(int): ID associated with the IDV attempt
        user_id(int): ID associated with the IDV attempt's user
        status(str): status in IDV language for the IDV attempt
        photo_id_name(str): name to be used as verified name
        full_name(str): user's pending name change or current profile name
    """
    trigger_status = VerifiedNameStatus.trigger_state_change_from_idv(status)

    # only trigger celery task if status is relevant to name affirmation
    if trigger_status:
        log.info('VerifiedName: idv_attempt_handler triggering Celery task for user %(user_id)s '
                 'with photo_id_name %(photo_id_name)s and status %(status)s',
                 {
                     'user_id': user_id,
                     'photo_id_name': photo_id_name,
                     'status': status
                 }
                 )
        idv_update_verified_name_task.delay(attempt_id, user_id, trigger_status, photo_id_name, full_name)
    else:
        log.info('VerifiedName: idv_attempt_handler will not trigger Celery task for user %(user_id)s '
                 'with photo_id_name %(photo_id_name)s because of status %(status)s',
                 {
                     'user_id': user_id,
                     'photo_id_name': photo_id_name,
                     'status': status
                 }
                 )


def idv_delete_handler(sender, instance, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Receiver for IDV attempt deletions

    Args:
        attempt_id(int): ID associated with the IDV attempt
    """
    idv_attempt_id = instance.id
    log.info(
        'VerifiedName: idv_delete_handler triggering Celery task for idv_attempt_id=%(idv_attempt_id)s',
        {
            'idv_attempt_id': idv_attempt_id,
        }
    )
    delete_verified_name_task.delay(idv_attempt_id, None)


def proctoring_attempt_handler(
    attempt_id,
    user_id,
    status,
    full_name,
    profile_name,
    is_practice_exam,
    is_proctored,
    backend_supports_onboarding,
    **kwargs
):
    """
    Receiver for proctored exam attempt updates.

    Args:
        attempt_id(int): ID associated with the proctored exam attempt
        user_id(int): ID associated with the proctored exam attempt's user
        status(str): status in proctoring language for the proctored exam attempt
        full_name(str): name to be used as verified name
        profile_name(str): user's current profile name
        is_practice_exam(boolean): if the exam attempt is for a practice exam
        is_proctored(boolean): if the exam attempt is for a proctored exam
        backend_supports_onboarding(boolean): if the exam attempt is for an exam with a backend that supports onboarding
    """

    # We only care about updates from onboarding exams, or from non-practice proctored exams with a backend that
    # does not support onboarding. This is because those two event types are guaranteed to contain verification events,
    # whereas timed exams and proctored exams with a backend that does support onboarding are not guaranteed
    is_onboarding_exam = is_practice_exam and is_proctored and backend_supports_onboarding
    reviewable_proctored_exam = is_proctored and not is_practice_exam and not backend_supports_onboarding
    if not (is_onboarding_exam or reviewable_proctored_exam):
        return

    trigger_status = VerifiedNameStatus.trigger_state_change_from_proctoring(status)

    # only trigger celery task if status is relevant to name affirmation
    if trigger_status:
        proctoring_update_verified_name_task.delay(
            attempt_id,
            user_id,
            trigger_status,
            full_name,
            profile_name,
        )
    else:
        log.info('VerifiedName: proctoring_attempt_handler will not trigger Celery task for user %(user_id)s '
                 'with profile_name %(profile_name)s because of status %(status)s',
                 {
                     'user_id': user_id,
                     'profile_name': profile_name,
                     'status': status,
                 }
                 )


def proctoring_delete_handler(sender, instance, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Receiver for proctoring attempt deletions

    Args:
        attempt_id(int): ID associated with the proctoring attempt
    """
    proctoring_attempt_id = instance.id
    log.info(
        'VerifiedName: proctoring_delete_handler triggering Celery task for '
        'proctoring_attempt_id=%(proctoring_attempt_id)s',
        {
            'proctoring_attempt_id': proctoring_attempt_id,
        }
    )
    delete_verified_name_task.delay(None, proctoring_attempt_id)



# Generated by Django 2.2.24 on 2021-08-18 10:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('edx_name_affirmation', '0004_auto_20210816_0958'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='verifiedname',
            name='is_verified',
        ),
    ]



"""
Name Affirmation Service
"""

import types


class NameAffirmationService:
    """
    Service to expose Name Affirmation API methods
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """
        This is a class factory to make sure this is a Singleton.
        """
        if not cls._instance:
            cls._instance = super(NameAffirmationService, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        """
        Class initializer, which just inspects the libraries and exposes the same functions
        as a direct pass through.
        """
        # pylint: disable=import-outside-toplevel
        from edx_name_affirmation import api as edx_name_affirmation_api
        self._bind_to_module_functions(edx_name_affirmation_api)

    def _bind_to_module_functions(self, module):
        """
        Bind module functions.
        """
        for attr_name in dir(module):
            attr = getattr(module, attr_name, None)
            if isinstance(attr, types.FunctionType):
                if not hasattr(self, attr_name):
                    setattr(self, attr_name, attr)



"""
Tests for Name Affirmation tasks
"""

import ddt
from mock import patch

from django.contrib.auth import get_user_model
from django.test import TestCase

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.statuses import VerifiedNameStatus
from edx_name_affirmation.tasks import (
    delete_verified_name_task,
    idv_update_verified_name_task,
    proctoring_update_verified_name_task
)

User = get_user_model()


@ddt.ddt
class TaskTests(TestCase):
    """
    Tests for tasks.py
    """
    def setUp(self):
        self.user = User(username='tester', email='tester@test.com')
        self.user.save()
        self.verified_name_obj = VerifiedName(
          user=self.user, verified_name='Jonathan Doe', profile_name='Jon Doe',
        )
        self.verified_name_obj.save()
        self.idv_attempt_id = 1111111
        self.proctoring_attempt_id = 2222222

    @patch('edx_name_affirmation.tasks.idv_update_verified_name_task.retry')
    def test_idv_retry(self, mock_retry):
        idv_update_verified_name_task.delay(
            self.idv_attempt_id,
            # force an error with an invalid user ID
            99999,
            VerifiedNameStatus.SUBMITTED,
            self.verified_name_obj.verified_name,
            self.verified_name_obj.profile_name,
        )
        mock_retry.assert_called()

    @patch('edx_name_affirmation.tasks.proctoring_update_verified_name_task.retry')
    def test_proctoring_retry(self, mock_retry):
        proctoring_update_verified_name_task.delay(
            self.proctoring_attempt_id,
            # force an error with an invalid user ID
            99999,
            VerifiedNameStatus.PENDING,
            self.verified_name_obj.verified_name,
            self.verified_name_obj.profile_name,
        )
        mock_retry.assert_called()

    def test_idv_delete(self):
        """
        Assert that only relevant VerifiedNames are deleted for a given idv_attempt_id
        """
        # associated test object with an idv attempt
        self.verified_name_obj.verification_attempt_id = self.idv_attempt_id
        self.verified_name_obj.save()

        other_attempt_id = 123456

        # create another VerifiedName with the same idv attempt
        VerifiedName(
            user=self.user,
            verified_name='Jonathan X Doe',
            profile_name='Jon D',
            verification_attempt_id=self.idv_attempt_id
        ).save()

        # create VerifiedName not associated with idv attempt
        other_verified_name_obj = VerifiedName(
            user=self.user,
            verified_name='Jonathan X Doe',
            profile_name='Jon D',
            verification_attempt_id=other_attempt_id
        )
        other_verified_name_obj.save()

        delete_verified_name_task.delay(
            self.idv_attempt_id,
            None
        )

        # check that there is only VerifiedName object
        self.assertEqual(len(VerifiedName.objects.filter(verification_attempt_id=self.idv_attempt_id)), 0)
        self.assertEqual(len(VerifiedName.objects.filter(verification_attempt_id=other_attempt_id)), 1)

    def test_proctoring_delete(self):
        """
        Assert that only relevant VerifiedNames are deleted for a given proctoring_attempt_id
        """
        # associated test object with a proctoring attempt
        self.verified_name_obj.proctored_exam_attempt_id = self.proctoring_attempt_id
        self.verified_name_obj.save()

        other_attempt_id = 123456

        # create another VerifiedName with the same proctoring attempt
        VerifiedName(
            user=self.user,
            verified_name='Jonathan X Doe',
            profile_name='Jon D',
            proctored_exam_attempt_id=self.proctoring_attempt_id
        ).save()

        # create VerifiedName not associated with proctoring attempt
        other_verified_name_obj = VerifiedName(
            user=self.user,
            verified_name='Jonathan X Doe',
            profile_name='Jon D',
            proctored_exam_attempt_id=other_attempt_id
        )
        other_verified_name_obj.save()

        delete_verified_name_task.delay(
            None,
            self.proctoring_attempt_id
        )

        # check that there is only VerifiedName object
        self.assertEqual(len(VerifiedName.objects.filter(proctored_exam_attempt_id=self.proctoring_attempt_id)), 0)
        self.assertEqual(len(VerifiedName.objects.filter(proctored_exam_attempt_id=other_attempt_id)), 1)

    @ddt.data(
        (1234, 5678),
        (None, None)
    )
    @ddt.unpack
    @patch('logging.Logger.error')
    def test_incorrect_args_delete(self, idv_attempt_id, proctoring_attempt_id, mock_logger):
        """
        Assert that error log is called and that no VerifiedNames are deleted when incorrect args are passed to task
        """
        delete_verified_name_task.delay(
            idv_attempt_id,
            proctoring_attempt_id
        )
        mock_logger.assert_called()

    @patch('logging.Logger.info')
    def test_no_names_delete(self, mock_logger):
        delete_verified_name_task.delay(
            self.idv_attempt_id,
            None
        )
        mock_logger.assert_called_with(
            'No VerifiedNames deleted because no VerifiedNames were associated with the provided attempt ID.'
        )



"""
Tests for Name Affirmation signal handlers
"""

import ddt
from mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.test import TestCase

from edx_name_affirmation.handlers import (
    idv_attempt_handler,
    idv_delete_handler,
    proctoring_attempt_handler,
    proctoring_delete_handler
)
from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.statuses import VerifiedNameStatus

User = get_user_model()


class SignalTestCase(TestCase):
    """
    Test case for signals.py
    """

    def setUp(self):
        self.user = User(username='tester', email='tester@test.com')
        self.user.save()
        self.verified_name = 'Jonathan Smith'
        self.profile_name = 'Jon Smith'
        self.idv_attempt_id = 1111111
        self.proctoring_attempt_id = 2222222


@ddt.ddt
class PostSaveVerifiedNameTests(SignalTestCase):
    """
    Tests for the post_save handler on the VerifiedName model.
    """

    @ddt.data(
      (VerifiedNameStatus.PENDING, False),
      (VerifiedNameStatus.SUBMITTED, False),
      (VerifiedNameStatus.APPROVED, True),
      (VerifiedNameStatus.DENIED, False)
    )
    @ddt.unpack
    def test_post_save_verified_name_approved(self, status, should_send):
        """
        Test that VERIFIED_NAME_APPROVED should only send if the status is changed to approved.
        """
        with patch('edx_name_affirmation.signals.VERIFIED_NAME_APPROVED.send') as mock_signal:
            verified_name_obj = VerifiedName.objects.create(
                user=self.user,
                verified_name='Jonathan Doe',
                profile_name=self.profile_name,
                verification_attempt_id=self.idv_attempt_id
            )
            verified_name_obj.status = status
            verified_name_obj.save()

            self.assertEqual(mock_signal.called, should_send)
            if should_send:
                mock_signal.assert_called_with(
                    sender='name_affirmation', user_id=self.user.id, profile_name=self.profile_name
                )


@ddt.ddt
class IDVSignalTests(SignalTestCase):
    """
    Test for idv_attempt_handler
    """

    def test_idv_create_verified_name(self):
        """
        Test that if no verified name exists for the name or attempt id, create one
        """
        idv_attempt_handler(
            self.idv_attempt_id,
            self.user.id,
            'created',
            self.verified_name,
            self.profile_name
        )

        # make sure that verifiedname is created with relevant data
        verified_name = VerifiedName.objects.get(verification_attempt_id=self.idv_attempt_id)
        self.assertEqual(verified_name.status, VerifiedNameStatus.PENDING)
        self.assertEqual(verified_name.verification_attempt_id, self.idv_attempt_id)
        self.assertEqual(verified_name.verified_name, self.verified_name)
        self.assertEqual(verified_name.profile_name, self.profile_name)

    @ddt.data(
        ('created', VerifiedNameStatus.PENDING),
        ('submitted', VerifiedNameStatus.SUBMITTED),
        ('approved', VerifiedNameStatus.APPROVED),
        ('denied', VerifiedNameStatus.DENIED)
    )
    @ddt.unpack
    def test_idv_update_multiple_verified_names(self, idv_status, expected_status):
        """
        If a VerifiedName(s) for a user and verified name exist, ensure that it is updated properly
        """
        # create multiple VerifiedNames
        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
        )
        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
        )
        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
            verification_attempt_id=self.idv_attempt_id
        )

        idv_attempt_handler(
            self.idv_attempt_id,
            self.user.id,
            idv_status,
            self.verified_name,
            self.profile_name
        )

        # check that the attempt id and status have been updated for all three VerifiedNames
        self.assertEqual(len(VerifiedName.objects.filter(verification_attempt_id=self.idv_attempt_id)), 3)
        self.assertEqual(len(VerifiedName.objects.filter(status=expected_status)), 3)

    def test_idv_create_with_existing_verified_names(self):
        """
        Test that if a user attempts IDV again with the same name as previous attempts, we still create a new record
        """
        previous_id = 1234567

        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
            verification_attempt_id=previous_id,
            status='denied'
        )

        # create an IDV attempt with the same user and names as above, but change the attempt ID to a unique value
        idv_attempt_handler(
            self.idv_attempt_id,
            self.user.id,
            'submitted',
            self.verified_name,
            self.profile_name
        )

        verified_name = VerifiedName.objects.get(verification_attempt_id=self.idv_attempt_id)
        self.assertEqual(verified_name.status, VerifiedNameStatus.SUBMITTED)

        previous_name = VerifiedName.objects.get(verification_attempt_id=previous_id)
        self.assertEqual(previous_name.status, VerifiedNameStatus.DENIED)

    def test_idv_does_not_update_verified_name_by_proctoring(self):
        """
        If the idv handler is triggered, ensure that the idv attempt info does not update any verified name
        records that have a proctoring attempt id
        """
        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
            proctored_exam_attempt_id=self.proctoring_attempt_id,
            status=VerifiedNameStatus.DENIED
        )
        VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name
        )

        idv_attempt_handler(
            self.idv_attempt_id,
            self.user.id,
            'submitted',
            self.verified_name,
            self.profile_name
        )

        # check that the attempt id and status have only been updated for the record that does not have a proctored
        # exam attempt id
        self.assertEqual(len(VerifiedName.objects.filter(verification_attempt_id=self.idv_attempt_id)), 1)
        self.assertEqual(len(VerifiedName.objects.filter(status=VerifiedNameStatus.SUBMITTED)), 1)

    @ddt.data(
        ('created', VerifiedNameStatus.PENDING),
        ('submitted', VerifiedNameStatus.SUBMITTED),
        ('approved', VerifiedNameStatus.APPROVED),
        ('denied', VerifiedNameStatus.DENIED)
    )
    @ddt.unpack
    def test_idv_update_one_verified_name(self, idv_status, expected_status):
        """
        If a VerifiedName(s) for a user and verified name exist, ensure that it is updated properly
        """
        with patch('edx_name_affirmation.signals.VERIFIED_NAME_APPROVED.send') as mock_signal:
            VerifiedName.objects.create(
                user=self.user,
                verified_name=self.verified_name,
                profile_name=self.profile_name,
                verification_attempt_id=self.idv_attempt_id
            )

            idv_attempt_handler(
                self.idv_attempt_id,
                self.user.id,
                idv_status,
                self.verified_name,
                self.profile_name
            )

            # check that the attempt id and status have been updated for all three VerifiedNames
            self.assertEqual(len(VerifiedName.objects.filter(verification_attempt_id=self.idv_attempt_id)), 1)
            self.assertEqual(len(VerifiedName.objects.filter(status=expected_status)), 1)

            # If the status is approved, ensure that the post_save signal is called
            if expected_status == VerifiedNameStatus.APPROVED:
                mock_signal.assert_called()
            else:
                mock_signal.assert_not_called()

    @ddt.data(
        'ready',
        'must_retry',
    )
    @patch('edx_name_affirmation.tasks.idv_update_verified_name_task.delay')
    def test_idv_non_trigger_status(self, status, mock_task):
        """
        Test that a celery task is not triggered if a non-relevant status is received
        """
        idv_attempt_handler(
            self.idv_attempt_id,
            self.user.id,
            status,
            self.verified_name,
            self.profile_name
        )

        mock_task.assert_not_called()

    @patch('edx_name_affirmation.tasks.delete_verified_name_task.delay')
    def test_idv_delete_handler(self, mock_task):
        """
        Test that a celery task is triggered if an idv delete signal is received
        """
        mock_idv_object = MagicMock()
        mock_idv_object.id = 'abcdef'
        idv_delete_handler(
            {},
            mock_idv_object,
            '',
        )

        mock_task.assert_called_with(mock_idv_object.id, None)


@ddt.ddt
class ProctoringSignalTests(SignalTestCase):
    """
    Test for proctoring_attempt_handler
    """

    @ddt.data(
        ('created', VerifiedNameStatus.PENDING),
        ('submitted', VerifiedNameStatus.SUBMITTED),
        ('verified', VerifiedNameStatus.APPROVED),
        ('rejected', VerifiedNameStatus.DENIED),
        ('error', VerifiedNameStatus.DENIED),
    )
    @ddt.unpack
    def test_proctoring_update_status_for_attempt_id(self, proctoring_status, expected_status):
        """
        If a verified name with an attempt ID already exists, update the VerifiedName status
        """
        # create a verified name with an attempt id
        verified_name = VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
            proctored_exam_attempt_id=self.proctoring_attempt_id,
        )
        object_id = verified_name.id

        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            proctoring_status,
            self.verified_name,
            self.profile_name,
            True,
            True,
            True
        )
        # make sure that status on verified name is correct
        verified_name_query = VerifiedName.objects.filter(id=object_id)
        self.assertEqual(len(verified_name_query), 1)
        verified_name = verified_name_query.first()
        self.assertEqual(verified_name.status, expected_status)

    @ddt.data(
        ('verified', VerifiedNameStatus.APPROVED),
        ('rejected', VerifiedNameStatus.DENIED),
    )
    @ddt.unpack
    def test_proctoring_error_update_status(self, proctoring_status, expected_status):
        """
        If we receive a proctoring update with an error status, ensure that later status updates are handled as expected
        """

        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'error',
            self.verified_name,
            self.profile_name,
            True,
            True,
            True
        )

        verified_name = VerifiedName.objects.get(proctored_exam_attempt_id=self.proctoring_attempt_id)
        object_id = verified_name.id
        self.assertEqual(verified_name.status, VerifiedNameStatus.DENIED)

        # update status
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            proctoring_status,
            self.verified_name,
            self.profile_name,
            True,
            True,
            True
        )

        # ensure that status is updated for subsequent updates
        verified_name_query = VerifiedName.objects.filter(id=object_id)
        self.assertEqual(len(verified_name_query), 1)
        verified_name = verified_name_query.first()
        self.assertEqual(verified_name.status, expected_status)

    def test_proctoring_create_verified_name(self):
        """
        Test that if no verified name exists for the name or attempt id, create one
        """
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'created',
            self.verified_name,
            self.profile_name,
            True,
            True,
            True
        )

        # make sure that verifiedname is created with relevant data
        verified_name_query = VerifiedName.objects.filter(proctored_exam_attempt_id=self.proctoring_attempt_id)
        self.assertEqual(len(verified_name_query), 1)
        verified_name = verified_name_query.first()
        self.assertEqual(verified_name.status, VerifiedNameStatus.PENDING)

    def test_proctoring_does_not_create_name(self):
        """
        Test that if we receive a signal for an attempt id that we do not have a name for, we do not create a new
        record.
        """

        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'created',
            None,
            None,
            True,
            True,
            True
        )

        self.assertEqual(len(VerifiedName.objects.filter()), 0)

    @ddt.data(
        (False, False, False),
        (False, True, True),
        (True, True, False)
    )
    @ddt.unpack
    @patch('edx_name_affirmation.tasks.proctoring_update_verified_name_task.delay')
    def test_proctoring_does_not_trigger_celery_task(self, is_practice, is_proctored, supports_onboarding, mock_task):
        """
        Test that a celery task is not triggered if the exam does not contain an id verification event
        """
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'created',
            'John',
            'John',
            is_practice,
            is_proctored,
            supports_onboarding
        )

        mock_task.assert_not_called()

    @ddt.data(
        True,
        False
    )
    @patch('logging.Logger.warning')
    def test_proctoring_log_with_existing_approved_verified_name(self, should_names_differ, mock_logger):
        """
        Test that we log a warning when we receive a proctoring signal that has a different full_name
        than the existing approved verified name
        """
        verified_name = VerifiedName.objects.create(
            user=self.user,
            verified_name=self.verified_name,
            profile_name=self.profile_name,
            proctored_exam_attempt_id=self.proctoring_attempt_id,
            status=VerifiedNameStatus.APPROVED
        )

        additional_attempt_id = self.proctoring_attempt_id + 1
        proctoring_attempt_handler(
            additional_attempt_id,
            self.user.id,
            'created',
            ('John' if should_names_differ else self.verified_name),
            ('John' if should_names_differ else self.profile_name),
            True,
            True,
            True
        )

        log_str = (
            'Full name for proctored_exam_attempt_id={attempt_id} is not equal to the most recent verified '
            'name verified_name_id={verified_name_id}.'
        ).format(
            attempt_id=additional_attempt_id,
            verified_name_id=verified_name.id
        )

        self.assertEqual(len(VerifiedName.objects.filter()), 1)
        if should_names_differ:
            mock_logger.assert_called_with(log_str)
        else:
            # check that log is not called if the names do not differ
            with self.assertRaises(AssertionError):
                mock_logger.assert_called_with(log_str)

    @ddt.data(
        'download_software_clicked',
        'ready_to_start',
        'started',
        'ready_to_submit',
    )
    @patch('edx_name_affirmation.tasks.proctoring_update_verified_name_task.delay')
    def test_proctoring_non_trigger_status(self, status, mock_task):
        """
        Test that a celery task is not triggered if a non-relevant status is received
        """
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            status,
            self.verified_name,
            self.profile_name,
            True,
            True,
            True
        )

        mock_task.assert_not_called()

    @patch('edx_name_affirmation.tasks.delete_verified_name_task.delay')
    def test_proctoring_delete_handler(self, mock_task):
        """
        Test that a celery task is triggered if an idv delete signal is received
        """
        mock_proctoring_object = MagicMock()
        mock_proctoring_object.id = 'abcdef'
        proctoring_delete_handler(
            {},
            mock_proctoring_object,
            '',
        )

        mock_task.assert_called_with(None, mock_proctoring_object.id)

    def test_proctoring_multiple_approved(self):
        # create task for submitted exam
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'submitted',
            'John',
            'John',
            True,
            True,
            True
        )

        # create task for submitted on another exam
        other_attempt_id = self.proctoring_attempt_id + 1
        proctoring_attempt_handler(
            other_attempt_id,
            self.user.id,
            'submitted',
            'John',
            'John',
            True,
            True,
            True
        )

        self.assertEqual(len(VerifiedName.objects.filter()), 2)
        self.assertEqual(len(VerifiedName.objects.filter(status=VerifiedNameStatus.SUBMITTED)), 2)

        # create task for approved exam 1
        proctoring_attempt_handler(
            self.proctoring_attempt_id,
            self.user.id,
            'verified',
            'John',
            'John',
            True,
            True,
            True
        )

        # create task for approved exam 2
        proctoring_attempt_handler(
            other_attempt_id,
            self.user.id,
            'verified',
            'John',
            'John',
            True,
            True,
            True
        )

        self.assertEqual(len(VerifiedName.objects.filter()), 2)
        self.assertEqual(len(VerifiedName.objects.filter(status=VerifiedNameStatus.APPROVED)), 2)



"""
edx_name_affirmation Django application initialization.
"""

from django.apps import AppConfig


class EdxNameAffirmationConfig(AppConfig):
    """
    Configuration for the edx_name_affirmation Django application.
    """

    name = 'edx_name_affirmation'

    plugin_app = {
        'url_config': {
            'lms.djangoapp': {
                'namespace': 'edx_name_affirmation',
                'regex': '^api/',
                'relative_path': 'urls',
            }
        },
        'signals_config': {
            'lms.djangoapp': {
                'relative_path': 'handlers',
                'receivers': [
                    {
                        'receiver_func_name': 'idv_attempt_handler',
                        'signal_path': 'lms.djangoapps.verify_student.signals.idv_update_signal',
                    },
                    {
                        'receiver_func_name': 'idv_delete_handler',
                        'signal_path': 'django.db.models.signals.post_delete',
                        'sender_path': 'lms.djangoapps.verify_student.models.SoftwareSecurePhotoVerification',
                    },
                    {
                        'receiver_func_name': 'proctoring_attempt_handler',
                        'signal_path': 'edx_proctoring.signals.exam_attempt_status_signal',
                    },
                    {
                        'receiver_func_name': 'proctoring_delete_handler',
                        'signal_path': 'django.db.models.signals.post_delete',
                        'sender_path': 'edx_proctoring.models.ProctoredExamStudentAttempt',
                    },
                ],
            }
        }
    }

    def ready(self):
        """
        Connect signal handlers.
        """
        from . import handlers  # pylint: disable=import-outside-toplevel,unused-import



"""
Name Affirmation HTTP-based API endpoints
"""
from edx_api_doc_tools import path_parameter, query_parameter, schema
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import status as http_status
from rest_framework.authentication import SessionAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from django.contrib.auth import get_user_model

from edx_name_affirmation.api import (
    create_verified_name,
    create_verified_name_config,
    delete_verified_name,
    get_verified_name,
    get_verified_name_history,
    should_use_verified_name_for_certs,
    update_verified_name_status
)
from edx_name_affirmation.exceptions import (
    VerifiedNameAttemptIdNotGiven,
    VerifiedNameDoesNotExist,
    VerifiedNameMultipleAttemptIds
)
from edx_name_affirmation.serializers import (
    UpdateVerifiedNameSerializer,
    VerifiedNameConfigSerializer,
    VerifiedNameSerializer
)
from edx_name_affirmation.statuses import VerifiedNameStatus


class AuthenticatedAPIView(APIView):
    """
    Authenticate API View.
    """
    authentication_classes = (SessionAuthentication, JwtAuthentication)
    permission_classes = (IsAuthenticated,)


class VerifiedNameView(AuthenticatedAPIView):
    """
    Endpoint for a VerifiedName.

    Supports:
        HTTP POST: Creates a new VerifiedName.
        HTTP GET: Returns an existing VerifiedName (by username or requesting user)
        HTTP PATCH: Update the status of a VerifiedName
        HTTP DELETE: Delete a VerifiedName
    """
    @schema(
        parameters=[
            query_parameter('username', str, 'The username of which verified name records might be associated'),
        ],
        responses={
            200: 'The verified_name record associated with the username provided',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
            404: 'The verified_name record associated with the username provided does not exist.'
        },
    )
    def get(self, request):
        """
        Get most recent verified name for the request user or for the specified username
        For example: /edx_name_affirmation/v1/verified_name?username=jdoe
        Example response: {
            "username": "jdoe",
            "verified_name": "Jonathan Doe",
            "profile_name": "Jon Doe",
            "verification_attempt_id": 123,
            "proctored_exam_attempt_id": None,
            "status": "approved",
            "use_verified_name_for_certs": False,
        }
        """
        username = request.GET.get('username')
        if username and not request.user.is_staff:
            return Response(
                status=http_status.HTTP_403_FORBIDDEN,
                data={"detail": "Must be a Staff User to Perform this request."}
            )

        user = get_user_model().objects.get(username=username) if username else request.user
        verified_name = get_verified_name(user, is_verified=True)
        if verified_name is None:
            return Response(
                status=404,
                data={'detail': 'There is no verified name related to this user.'}
            )

        serialized_data = VerifiedNameSerializer(verified_name).data
        serialized_data['use_verified_name_for_certs'] = should_use_verified_name_for_certs(user)
        return Response(serialized_data)

    @schema(
        body=VerifiedNameSerializer(),
        responses={
            200: 'The verified_name record associated with the username provided is successfully created',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
            400: 'The posted data have conflicts with already stored verified name'
        },
    )
    def post(self, request):
        """
        Creates a new VerifiedName.
        Expected POST data: {
            "username": "jdoe",
            "verified_name": "Jonathan Doe"
            "profile_name": "Jon Doe"
            "verification_attempt_id": (Optional)
            "proctored_exam_attempt_id": (Optional)
            "status": (Optional)
        }
        """
        username = request.data.get('username')
        if username != request.user.username and not request.user.is_staff:
            return Response(
                status=http_status.HTTP_403_FORBIDDEN,
                data={"detail": "Must be a Staff User to Perform this request."}
            )

        serializer = VerifiedNameSerializer(data=request.data)
        if serializer.is_valid():
            user = get_user_model().objects.get(username=username) if username else request.user
            try:
                create_verified_name(
                    user,
                    request.data.get('verified_name'),
                    request.data.get('profile_name'),
                    verification_attempt_id=request.data.get('verification_attempt_id', None),
                    proctored_exam_attempt_id=request.data.get('proctored_exam_attempt_id', None),
                    status=request.data.get('status', VerifiedNameStatus.PENDING)
                )
                response_status = http_status.HTTP_200_OK
                data = {}
            except VerifiedNameMultipleAttemptIds as exc:
                response_status = http_status.HTTP_400_BAD_REQUEST
                data = {"detail": str(exc)}
        else:
            response_status = http_status.HTTP_400_BAD_REQUEST
            data = serializer.errors
        return Response(status=response_status, data=data)

    @schema(
        body=UpdateVerifiedNameSerializer(),
        responses={
            200: 'The verified_name record associated with the username provided is successfully edited',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
            400: 'The edit action failed validation rules'
        },
    )
    def patch(self, request):
        """
        Update verified name status
        Example PATCH data: {
                "username": "jdoe",
                "verification_attempt_id" OR "proctored_exam_attempt_id": 123,
                "status": "approved",
            }
        """
        if not request.user.is_staff:
            return Response(
                status=http_status.HTTP_403_FORBIDDEN,
                data={'detail': 'Must be a staff user to update verified name status.'}
            )

        serializer = UpdateVerifiedNameSerializer(data=request.data)

        if serializer.is_valid():
            username = request.data.get('username')
            user = get_user_model().objects.get(username=username)
            try:
                verified_name = update_verified_name_status(
                    user,
                    request.data.get('status'),
                    request.data.get('verification_attempt_id', None),
                    request.data.get('proctored_exam_attempt_id', None)
                )
                response_status = http_status.HTTP_200_OK
                data = VerifiedNameSerializer(verified_name).data
            except (VerifiedNameAttemptIdNotGiven, VerifiedNameMultipleAttemptIds) as exc:
                response_status = http_status.HTTP_400_BAD_REQUEST
                data = {'detail': str(exc)}
            except VerifiedNameDoesNotExist as exc:
                response_status = http_status.HTTP_404_NOT_FOUND
                data = {'detail': str(exc)}
        else:
            response_status = http_status.HTTP_400_BAD_REQUEST
            data = serializer.errors

        return Response(status=response_status, data=data)

    @schema(
        parameters=[
            path_parameter('verified_name_id', str, 'The database id of the verified_name to be deleted'),
        ],
        responses={
            204: 'The verified_name record associated with the id is successfully deleted from data store',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
            404: 'The verified_name record associated with the id provided does not exist.'
        },
    )
    def delete(self, request, verified_name_id):
        """
        Delete verified name
        /edx_name_affirmation/v1/verified_name/{verified_name_id}
        """
        if not request.user.is_staff:
            return Response(
                status=http_status.HTTP_403_FORBIDDEN,
                data={'detail': 'Must be a staff user to delete a verified name.'}
            )

        try:
            delete_verified_name(verified_name_id)
            response_status = http_status.HTTP_204_NO_CONTENT
            data = {}
        except VerifiedNameDoesNotExist as exc:
            response_status = http_status.HTTP_404_NOT_FOUND
            data = {'detail': str(exc)}

        return Response(status=response_status, data=data)


class VerifiedNameHistoryView(AuthenticatedAPIView):
    """
    Endpoint for VerifiedName history.
    /edx_name_affirmation/v1/verified_name/history?username=xxx

    Supports:
        HTTP GET: Return a list of VerifiedNames for the given user.
    """
    @schema(
        parameters=[
            query_parameter('username', str, 'The username of which verified name records might be associated'),
        ],
        responses={
            200: 'The verified_name record associated with the username provided is successfully edited',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
        },
    )
    def get(self, request):
        """
        Get a list of verified name objects for the given user, ordered by most recently created.
        """
        username = request.GET.get('username')
        if username and not request.user.is_staff:
            return Response(
                status=http_status.HTTP_403_FORBIDDEN,
                data={"detail": "Must be a Staff User to Perform this request."}
            )

        user = get_user_model().objects.get(username=username) if username else request.user
        verified_name_qs = get_verified_name_history(user)
        serializer = VerifiedNameSerializer(verified_name_qs, many=True)

        serialized_data = {
            'use_verified_name_for_certs': should_use_verified_name_for_certs(user),
            'results': serializer.data,
        }

        return Response(serialized_data)


class VerifiedNameConfigView(AuthenticatedAPIView):
    """
    Endpoint for VerifiedNameConfig.
    /edx_name_affirmation/v1/verified_name/config

    Supports:
        HTTP POST: Creates a new VerifiedNameConfig.

    HTTP POST
    Creates a new VerifiedName.
    Example POST data: {
        "username": "jdoe",
        "use_verified_name_for_certs": True
    }
    """
    @schema(
        body=VerifiedNameConfigSerializer(),
        responses={
            201: 'The verified_name configuration record is successfully created',
            403: 'User lacks required permission. Only an edX staff user can invoke this API',
            400: 'The POSTed data failed validation rules',
        },
    )
    def post(self, request):
        """
        Create VerifiedNameConfig
        """
        username = request.data.get('username')
        if username != request.user.username and not request.user.is_staff:
            msg = 'Must be a staff user to override the requested user’s VerifiedNameConfig value'
            return Response(status=http_status.HTTP_403_FORBIDDEN, data={'detail': msg})

        serializer = VerifiedNameConfigSerializer(data=request.data)

        if serializer.is_valid():
            user = get_user_model().objects.get(username=username) if username else request.user
            create_verified_name_config(
                user,
                use_verified_name_for_certs=request.data.get('use_verified_name_for_certs'),
            )
            response_status = http_status.HTTP_201_CREATED
            data = {}

        else:
            response_status = http_status.HTTP_400_BAD_REQUEST
            data = serializer.errors

        return Response(status=response_status, data=data)



"""
Tests for name change validator
"""


import ddt

from django.test import TestCase

from edx_name_affirmation.name_change_validator import NameChangeValidator


@ddt.ddt
class NameChangeValidatorTests(TestCase):
    """
    Tests for name_change_validator.py
    """

    @ddt.data(
        ('Jonathan Adams', 'Jon Adams'),
        ('Jonathan Adams', 'Jonathan Quincy Adams'),
        ('Jonathan Adams', 'Jon at han Adams'),
        ('Jonathan Adams', 'Jonathan  Adams'),
        ('Jonathan Adams', 'Jonathan Adens')
    )
    @ddt.unpack
    def test_name_update_requires_idv_invalid_edits(self, old_name, new_name):
        """
        Test that a name change is blocked through this API if it requires ID verification.
        In this case, the user has invalid name edits
        """
        validator = NameChangeValidator([], 1, old_name, new_name)
        self.assertFalse(validator.validate())

    def test_name_update_requires_idv_name_changes(self):
        """
        Test that a name change is blocked through this API if it requires ID verification.
        In this case, the user has previously changed their name 2 or more times
        """
        validator = NameChangeValidator(['Old Name 1'], 1, 'Old Name 2', 'New Name')
        self.assertFalse(validator.validate())

    @ddt.data(
        ('Jonathan Adams', 'Jonathan Q Adams'),
        ('Jonathan Adams', 'Jonathan Adam'),
        ('Jonathan Adams', 'Jo nathan Adams'),
        ('Jonathan Adams', 'Jonatha N Adams'),
        ('Jonathan Adams', 'Jonathan Adáms'),
        ('Jonathan Adáms', 'Jonathan Adæms'),
        ('Jonathan Adams', 'Jonathan A\'dams'),
        ('李陈', '李王')
    )
    @ddt.unpack
    def test_name_update_does_not_require_idv_valid_edits(self, old_name, new_name):
        """
        Test that the user can change their name freely if it does not require verification.
        """
        validator = NameChangeValidator([], 1, old_name, new_name)
        self.assertTrue(validator.validate())

    def test_name_update_does_not_require_idv_no_certificate(self):
        """
        Test that the user can change their name freely if they have no certificates
        """
        validator = NameChangeValidator(['Really Old Name', 'Very Old Name'], 0, 'Old Name', 'New Name')
        self.assertTrue(validator.validate())



"""
URLs for edx_name_affirmation.
"""

from django.urls import include, path, re_path

from edx_name_affirmation import views

app_name = 'edx_name_affirmation'

urlpatterns = [
    path(
        'edx_name_affirmation/v1/verified_name', views.VerifiedNameView.as_view(),
        name='verified_name'
    ),

    re_path(
        r'edx_name_affirmation/v1/verified_name/(?P<verified_name_id>\d+)$',
        views.VerifiedNameView.as_view(), name='verified_name_by_id'
    ),

    path(
        'edx_name_affirmation/v1/verified_name/history', views.VerifiedNameHistoryView.as_view(),
        name='verified_name_history'
    ),

    path(
        'edx_name_affirmation/v1/verified_name/config', views.VerifiedNameConfigView.as_view(),
        name='verified_name_config'
    ),

    path('', include('rest_framework.urls', namespace='rest_framework')),
]



"""
Custom Django admin pages for Name Affirmation
"""

from django.contrib import admin

from edx_name_affirmation.models import VerifiedName, VerifiedNameConfig


@admin.register(VerifiedName)
class VerifiedNameAdmin(admin.ModelAdmin):
    """
    Admin for the VerifiedName Model
    """
    list_display = (
      'id', 'user', 'verified_name', 'verification_attempt_id', 'proctored_exam_attempt_id',
      'status', 'created', 'modified',
    )
    readonly_fields = ('id',)
    search_fields = ('user__username', 'verification_attempt_id', 'proctored_exam_attempt_id',)
    raw_id_fields = ('user', )


@admin.register(VerifiedNameConfig)
class VerifiedNameConfigAdmin(admin.ModelAdmin):
    """
    Admin for the VerifiedNameConfig Model
    """
    list_display = (
      'id', 'user', 'use_verified_name_for_certs', 'change_date',
    )
    readonly_fields = ('change_date',)
    search_fields = ('user__username',)
    raw_id_fields = ('user', )



# Generated by Django 2.2.24 on 2021-06-15 09:05

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='VerifiedName',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('verified_name', models.CharField(db_index=True, max_length=255)),
                ('profile_name', models.CharField(max_length=255, null=True)),
                ('verification_attempt_id', models.PositiveIntegerField(null=True)),
                ('proctored_exam_attempt_id', models.PositiveIntegerField(null=True)),
                ('is_verified', models.BooleanField(default=False)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'verified name',
                'db_table': 'nameaffirmation_verifiedname',
            },
        ),
    ]



"""
All tests for edx_name_affirmation views
"""
import json
from unittest.mock import PropertyMock, patch

import ddt

from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.urls import reverse

from edx_name_affirmation.api import (
    create_verified_name,
    create_verified_name_config,
    get_verified_name,
    get_verified_name_history,
    should_use_verified_name_for_certs
)
from edx_name_affirmation.models import VerifiedNameConfig
from edx_name_affirmation.statuses import VerifiedNameStatus

from .utils import LoggedInTestCase

User = get_user_model()


@ddt.ddt
class NameAffirmationViewsTestCase(LoggedInTestCase):
    """
    Base test class for Name Affirmation views
    """

    def setUp(self):
        super().setUp()
        self.other_user = User(username='other_tester', email='other@test.com')
        self.other_user.save()
        # Create fresh configs with default values
        VerifiedNameConfig.objects.create(user=self.user)
        VerifiedNameConfig.objects.create(user=self.other_user)

    def tearDown(self):
        super().tearDown()
        cache.clear()


@ddt.ddt
class VerifiedNameViewTests(NameAffirmationViewsTestCase):
    """
    Tests for the VerifiedNameView
    """

    VERIFIED_NAME = 'Jonathan Doe'
    PROFILE_NAME = 'Jon Doe'

    OTHER_VERIFIED_NAME = 'Robert Smith'
    OTHER_PROFILE_NAME = 'Bob Smith'

    ATTEMPT_ID = 11111

    def test_verified_name(self):
        verified_name = self._create_verified_name(self.user, status=VerifiedNameStatus.APPROVED)

        expected_data = self._get_expected_data(self.user, verified_name)

        response = self.client.get(reverse('edx_name_affirmation:verified_name'))
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))
        self.assertEqual(data, expected_data)

    def test_verified_name_existing_config(self):
        verified_name = self._create_verified_name(self.user)
        create_verified_name_config(self.user, use_verified_name_for_certs=True)
        expected_data = self._get_expected_data(self.user, verified_name, use_verified_name_for_certs=True)

        response = self.client.get(reverse('edx_name_affirmation:verified_name'))
        data = json.loads(response.content.decode('utf-8'))
        self.assertTrue(data, expected_data)

    def test_staff_access_verified_name(self):
        create_verified_name(self.other_user, self.VERIFIED_NAME, self.PROFILE_NAME, status=VerifiedNameStatus.APPROVED)

        # check that non staff access returns 403
        response = self.client.get(
            reverse('edx_name_affirmation:verified_name'),
            {'username': self.other_user.username}
        )
        self.assertEqual(response.status_code, 403)

        self.user.is_staff = True
        self.user.save()

        # create verified name
        self._create_verified_name(self.user)
        other_user_verified_name = get_verified_name(self.other_user, is_verified=True)

        # expected data should match the verifiedname from the other user
        expected_data = self._get_expected_data(self.other_user, other_user_verified_name)

        response = self.client.get(
            reverse('edx_name_affirmation:verified_name'),
            {'username': self.other_user.username}
        )
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))
        self.assertEqual(data, expected_data)

    def test_404_if_no_verified_name(self):
        response = self.client.get(reverse('edx_name_affirmation:verified_name'))
        self.assertEqual(response.status_code, 404)

    def test_post_200(self):
        verified_name_data = {
            'username': self.user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': self.VERIFIED_NAME,
            'verification_attempt_id': self.ATTEMPT_ID,
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 200)

        created_name = get_verified_name(self.user, is_verified=False)
        self.assertEqual(created_name.user.username, self.user.username)
        self.assertEqual(created_name.profile_name, self.PROFILE_NAME)
        self.assertEqual(created_name.verified_name, self.VERIFIED_NAME)
        self.assertEqual(created_name.verification_attempt_id, self.ATTEMPT_ID)

    def test_post_200_if_staff(self):
        self.user.is_staff = True
        self.user.save()

        verified_name_data = {
            'username': self.other_user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': self.VERIFIED_NAME,
            'proctored_exam_attempt_id': self.ATTEMPT_ID,
            'status': VerifiedNameStatus.APPROVED.value,
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 200)

        created_name = get_verified_name(self.other_user, is_verified=True)
        self.assertEqual(created_name.user.username, self.other_user.username)
        self.assertEqual(created_name.profile_name, self.PROFILE_NAME)
        self.assertEqual(created_name.verified_name, self.VERIFIED_NAME)
        self.assertEqual(created_name.proctored_exam_attempt_id, self.ATTEMPT_ID)

    def test_post_403_non_staff(self):
        verified_name_data = {
            'username': self.other_user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': self.VERIFIED_NAME,
            'verification_attempt_id': self.ATTEMPT_ID,
            'status': VerifiedNameStatus.APPROVED.value,
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 403)

    @ddt.data('<html>Verified Name</html>', 'https://verifiedname.com')
    def test_post_400_invalid_name(self, verified_name):
        verified_name_data = {
            'username': self.user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': verified_name,
            'verification_attempt_id': self.ATTEMPT_ID,
            'status': VerifiedNameStatus.SUBMITTED.value,
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 400)

    def test_post_400_invalid_serializer(self):
        verified_name_data = {
            'username': self.user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': self.VERIFIED_NAME,
            'verification_attempt_id': 'xxyz',
            'status': VerifiedNameStatus.APPROVED.value,
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 400)

    def test_post_400_two_attempt_ids(self):
        verified_name_data = {
            'username': self.user.username,
            'profile_name': self.PROFILE_NAME,
            'verified_name': self.VERIFIED_NAME,
            'verification_attempt_id': self.ATTEMPT_ID,
            'proctored_exam_attempt_id': self.ATTEMPT_ID
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name'),
            verified_name_data
        )
        self.assertEqual(response.status_code, 400)

    @ddt.data((True, True, 200), (False, True, 403), (False, False, 403), (True, False, 404))
    @ddt.unpack
    def test_patch(self, is_staff, verified_name_exists, expected_response):
        if is_staff:
            self.user.is_staff = True
            self.user.save()

        if verified_name_exists:
            self._create_verified_name(
                self.other_user,
                verification_attempt_id=123,
                status=VerifiedNameStatus.SUBMITTED
            )

        update_data = {
            'username': self.other_user.username,
            'verification_attempt_id': 123,
            'status': VerifiedNameStatus.APPROVED
        }

        response = self.client.patch(
            reverse('edx_name_affirmation:verified_name'),
            update_data,
            content_type='application/json'
        )

        self.assertEqual(response.status_code, expected_response)
        if verified_name_exists:
            updated_verified_name = get_verified_name(self.other_user)
            if is_staff:
                data = json.loads(response.content.decode('utf-8'))
                assert updated_verified_name.status == data['status'] == VerifiedNameStatus.APPROVED
            else:
                self.assertEqual(updated_verified_name.status, VerifiedNameStatus.SUBMITTED)

    def test_patch_missing_status(self):
        self.user.is_staff = True
        self.user.save()

        self._create_verified_name(self.other_user, verification_attempt_id=123)

        update_data = {
            'username': self.other_user.username,
            'verification_attempt_id': 123
        }

        response = self.client.patch(
            reverse('edx_name_affirmation:verified_name'),
            update_data,
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)

    def test_patch_two_attempt_ids(self):
        self.user.is_staff = True
        self.user.save()

        self._create_verified_name(self.other_user, verification_attempt_id=123)

        update_data = {
            'username': self.other_user.username,
            'status': VerifiedNameStatus.APPROVED,
            'verification_attempt_id': 123,
            'proctored_exam_attempt_id': 456
        }

        response = self.client.patch(
            reverse('edx_name_affirmation:verified_name'),
            update_data,
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)

    def test_patch_missing_attempt_id(self):
        self.user.is_staff = True
        self.user.save()

        self._create_verified_name(self.other_user, verification_attempt_id=123)

        update_data = {
            'username': self.other_user.username,
            'status': VerifiedNameStatus.APPROVED
        }

        response = self.client.patch(
            reverse('edx_name_affirmation:verified_name'),
            update_data,
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 400)

    @ddt.data((True, True, 204), (False, True, 403), (False, False, 403), (True, False, 404))
    @ddt.unpack
    def test_delete(self, is_staff, verified_name_exists, expected_response):
        if is_staff:
            self.user.is_staff = True
            self.user.save()

        verified_name_id = 1
        if verified_name_exists:
            verified_name = self._create_verified_name(
                self.other_user,
                verification_attempt_id=123,
                status=VerifiedNameStatus.APPROVED
            )
            verified_name_id = verified_name.id

        response = self.client.delete(reverse(
            'edx_name_affirmation:verified_name_by_id',
            kwargs={'verified_name_id': verified_name_id}
        ))

        self.assertEqual(response.status_code, expected_response)
        if verified_name_exists:
            if is_staff:
                self.assertIsNone(get_verified_name(self.other_user))
            else:
                self.assertEqual(get_verified_name(self.other_user), verified_name)

    def _create_verified_name(
        self, user, verification_attempt_id=None,
        proctored_exam_attempt_id=None, status=VerifiedNameStatus.PENDING,
    ):
        """
        Create and return a verified name object.
        """
        create_verified_name(
            user,
            self.VERIFIED_NAME,
            self.PROFILE_NAME,
            verification_attempt_id,
            proctored_exam_attempt_id,
            status
        )
        return get_verified_name(user)

    def _get_expected_data(
        self, user, verified_name_obj,
        use_verified_name_for_certs=False,
    ):
        """
        Create a dictionary of expected data.
        """
        return {
            'id': verified_name_obj.id,
            'created': verified_name_obj.created.isoformat(),
            'username': user.username,
            'verified_name': verified_name_obj.verified_name,
            'profile_name': verified_name_obj.profile_name,
            'verification_attempt_id': verified_name_obj.verification_attempt_id,
            'verification_attempt_status': None,
            'proctored_exam_attempt_id': verified_name_obj.proctored_exam_attempt_id,
            'status': verified_name_obj.status,
            'use_verified_name_for_certs': use_verified_name_for_certs,
        }


@ddt.ddt
class VerifiedNameHistoryViewTests(NameAffirmationViewsTestCase):
    """
    Tests for the VerifiedNameHistoryView
    """

    def test_get(self):
        verified_name_history = self._create_verified_name_history(self.user)
        expected_response = self._get_expected_response(self.user, verified_name_history)

        response = self.client.get(reverse('edx_name_affirmation:verified_name_history'))

        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))
        self.assertEqual(data, expected_response)

    @patch('edx_name_affirmation.api.VerifiedName.verification_attempt_status', new_callable=PropertyMock)
    def test_get_with_idv_status(self, verification_attempt_status_mock):
        mocked_idv_status = 'approved'

        verification_attempt_status_mock.return_value = mocked_idv_status
        verified_name_history = self._create_verified_name_history(self.user)
        expected_response = self._get_expected_response(self.user, verified_name_history)

        # replacing the expected response results with the mocked status
        for row in expected_response['results']:
            row['verification_attempt_status'] = mocked_idv_status

        response = self.client.get(reverse('edx_name_affirmation:verified_name_history'))

        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))

        self.assertEqual(data, expected_response)

    def test_get_bools(self):
        verified_name_history = self._create_verified_name_history(self.user)
        expected_response = self._get_expected_response(
            self.user, verified_name_history,
            use_verified_name_for_certs=False
        )

        response = self.client.get(reverse('edx_name_affirmation:verified_name_history'))

        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))
        self.assertEqual(data, expected_response)

    def test_get_no_data(self):
        expected_response = self._get_expected_response(self.user, [])
        response = self.client.get(reverse('edx_name_affirmation:verified_name_history'))

        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf-8'))
        self.assertEqual(data, expected_response)

    @ddt.data((True, 200), (False, 403))
    @ddt.unpack
    def test_get_staff_access(self, is_staff, expected_response):
        if is_staff:
            self.user.is_staff = True
            self.user.save()

        response = self.client.get(
            reverse('edx_name_affirmation:verified_name_history'),
            {'username': self.other_user.username}
        )

        self.assertEqual(response.status_code, expected_response)

    def _create_verified_name_history(self, user):
        """
        Create and return a verified name QuerySet.
        """
        create_verified_name(
            user,
            'Jonathan Doe',
            'Jon Doe',
            verification_attempt_id=123,
            status=VerifiedNameStatus.APPROVED,
        )
        create_verified_name(
            user,
            'Jane Doe',
            'Jane Doe',
            proctored_exam_attempt_id=456,
            status=VerifiedNameStatus.DENIED,
        )
        return get_verified_name_history(user)

    def _get_expected_response(
        self,
        user,
        verified_name_history,
        use_verified_name_for_certs=False
    ):
        """
        Create and return a verified name QuerySet.
        """
        expected_response = {
            'results': [],
            'use_verified_name_for_certs': use_verified_name_for_certs,
        }

        for verified_name_obj in verified_name_history:
            data = {
                'id': verified_name_obj.id,
                'created': verified_name_obj.created.isoformat(),
                'username': user.username,
                'verified_name': verified_name_obj.verified_name,
                'profile_name': verified_name_obj.profile_name,
                'verification_attempt_id': verified_name_obj.verification_attempt_id,
                'verification_attempt_status': None,
                'proctored_exam_attempt_id': verified_name_obj.proctored_exam_attempt_id,
                'status': verified_name_obj.status
            }
            expected_response['results'].append(data)

        return expected_response


class VerifiedNameConfigViewTests(NameAffirmationViewsTestCase):
    """
    Tests for the VerifiedNameConfigView
    """

    def test_post_201(self):
        config_data = {
            'username': self.user.username,
            'use_verified_name_for_certs': True
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            config_data
        )
        self.assertEqual(response.status_code, 201)

        use_verified_name_for_certs = should_use_verified_name_for_certs(self.user)
        self.assertTrue(use_verified_name_for_certs)

    def test_post_201_missing_field(self):
        initial_config_data = {
            'username': self.user.username,
            'use_verified_name_for_certs': True
        }
        config_data_missing_field = {'username': self.user.username}

        first_response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            initial_config_data
        )
        second_response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            config_data_missing_field
        )

        self.assertEqual(first_response.status_code, 201)
        self.assertEqual(second_response.status_code, 201)

        # `use_verified_name_for_certs` should not be overriden with False due to a missing field
        use_verified_name_for_certs = should_use_verified_name_for_certs(self.user)
        self.assertTrue(use_verified_name_for_certs)

    def test_post_201_if_staff(self):
        self.user.is_staff = True
        self.user.save()

        config_data = {
            'username': self.other_user.username,
            'use_verified_name_for_certs': True
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            config_data
        )
        self.assertEqual(response.status_code, 201)

        use_verified_name_for_certs = should_use_verified_name_for_certs(self.other_user)
        self.assertTrue(use_verified_name_for_certs)

    def test_post_403_non_staff(self):
        config_data = {
            'username': self.other_user.username,
            'use_verified_name_for_certs': True
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            config_data
        )
        self.assertEqual(response.status_code, 403)

    def test_post_400_invalid_serializer(self):
        config_data = {
            'username': self.user.username,
            'use_verified_name_for_certs': 'not a boolean'
        }
        response = self.client.post(
            reverse('edx_name_affirmation:verified_name_config'),
            config_data
        )
        self.assertEqual(response.status_code, 400)



"""
Statuses for edx_name_affirmation.
"""

from enum import Enum


class VerifiedNameStatus(str, Enum):
    """
    Possible states for the verified name.

    Pending: the verified name has been created

    Submitted: the verified name has been submitted to a verification authority

    Approved, Denied: resulting states from that authority

    This is the status of the verified name attempt, which is related to
    but separate from the status of the verifying process such as IDV or proctoring.
    Status changes in the verifying processes are usually more fine grained.

    For example when proctoring changes from ready to start to started the verified
    name is still pending. Once proctoring is actually submitted the verified name
    can be considered submitted.

    The expected lifecycle is pending -> submitted -> approved/denied.

    .. no_pii: This model has no PII.
    """
    PENDING = "pending"
    SUBMITTED = "submitted"
    APPROVED = "approved"
    DENIED = "denied"

    @classmethod
    def trigger_state_change_from_idv(cls, idv_status):
        """
        Return the translated IDV status if it should trigger a state transition, otherwise return None
        """
        # mapping from an idv status (key) to it's associated verified name status (value). We only want to
        # include idv statuses that would cause a status transition for a verified name
        idv_state_transition_mapping = {
            'created': cls.PENDING,
            'submitted': cls.SUBMITTED,
            'approved': cls.APPROVED,
            'denied': cls.DENIED
        }

        return idv_state_transition_mapping.get(idv_status, None)

    @classmethod
    def trigger_state_change_from_proctoring(cls, proctoring_status):
        """
        Return the translated proctoring status if it should trigger a state transition, otherwise return None
        """
        # mapping from an proctoring status (key) to it's associated verified name status (value). We only want to
        # include proctoring statuses that would cause a status transition for a verified name
        proctoring_state_transition_mapping = {
            'created': cls.PENDING,
            'submitted': cls.SUBMITTED,
            'verified': cls.APPROVED,
            'rejected': cls.DENIED,
            'error': cls.DENIED,
        }

        return proctoring_state_transition_mapping.get(proctoring_status, None)



# Generated by Django 3.2.5 on 2021-07-19 09:24

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('edx_name_affirmation', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='VerifiedNameConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('change_date', models.DateTimeField(auto_now_add=True, verbose_name='Change date')),
                ('enabled', models.BooleanField(default=False, verbose_name='Enabled')),
                ('use_verified_name_for_certs', models.BooleanField(default=False)),
                ('changed_by', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='Changed by')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='verified_name_config', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'verified name config',
                'db_table': 'nameaffirmation_verifiednameconfig',
            },
        ),
    ]



"""Subclasses Django test client to allow for easy login"""

from importlib import import_module

from django.conf import settings
from django.contrib.auth import get_user_model, login
from django.http import HttpRequest
from django.test import TestCase
from django.test.client import Client

User = get_user_model()


class TestClient(Client):
    """
    Allows for 'fake logins' of a user so we don't need to expose a 'login' HTTP endpoint
    """
    def login_user(self, user):
        """
        Login as specified user, does not depend on auth backend (hopefully)

        This is based on Client.login() with a small hack that does not
        require the call to authenticate()
        """
        user.backend = "django.contrib.auth.backends.ModelBackend"
        engine = import_module(settings.SESSION_ENGINE)

        # Create a fake request to store login details.
        request = HttpRequest()

        request.session = engine.SessionStore()
        login(request, user)

        # Set the cookie to represent the session.
        session_cookie = settings.SESSION_COOKIE_NAME
        self.cookies[session_cookie] = request.session.session_key
        cookie_data = {
            'max-age': None,
            'path': '/',
            'domain': settings.SESSION_COOKIE_DOMAIN,
            'secure': settings.SESSION_COOKIE_SECURE or None,
            'expires': None,
        }
        self.cookies[session_cookie].update(cookie_data)

        # Save the session values.
        request.session.save()


class LoggedInTestCase(TestCase):
    """
    All tests for the views.py
    """

    def setUp(self):
        """
        Setup for tests
        """
        super().setUp()
        self.client = TestClient()
        self.user = User(username='tester', email='tester@test.com')
        self.user.save()
        self.client.login_user(self.user)



"""
Tests for NameAffirmation services
"""

import types

from django.contrib.auth import get_user_model
from django.test import TestCase

from edx_name_affirmation import api as edx_name_affirmation_api
from edx_name_affirmation.services import NameAffirmationService

User = get_user_model()


class NameAffirmationServiceTest(TestCase):
    """
    Tests for NameAffirmationService
    """
    def test_basic(self):
        """
        See if the NameAffirmationService exposes the expected methods
        """

        service = NameAffirmationService()

        for attr_name in dir(edx_name_affirmation_api):
            attr = getattr(edx_name_affirmation_api, attr_name, None)
            if isinstance(attr, types.FunctionType):
                self.assertTrue(hasattr(service, attr_name))

    def test_singleton(self):
        """
        Test to make sure the NameAffirmationService is a singleton.
        """
        service1 = NameAffirmationService()
        service2 = NameAffirmationService()
        self.assertIs(service1, service2)
